"""
Python SDK service layer
"""

from kili.services.asset_import import import_assets
from kili.services.export import export_labels
from kili.services.label_import import import_labels_from_dict, import_labels_from_files
from kili.services.project import get_project, get_project_field

__all__ = [
    "export_labels",
    "import_assets",
    "import_labels_from_files",
    "import_labels_from_dict",
    "get_project",
    "get_project_field",
]



"""CLI's project import subcommand"""

import os
import urllib.request
from typing import Iterable, Optional

import click
from typeguard import typechecked

from kili import services
from kili.cli.common_args import Arguments, Options, from_csv
from kili.cli.helpers import collect_from_csv, get_kili_client
from kili.helpers import get_file_paths_to_upload
from kili.services import asset_import
from kili.services.helpers import (
    check_exclusive_options,
    get_external_id_from_file_path,
)


def check_asset_type(key, value):
    """type check value based on key"""
    if key == "content" and not os.path.isfile(value):
        with urllib.request.urlopen(value) as http_response:
            if not http_response.getcode() == 200:
                return f"{value} is not a valid url or path to a file."

    return ""


def generate_json_metadata(as_frames, fps):
    """Generate the json_metadata for input of the import_assets service
    when uploading from a list of path

    Args:
        as_frames: for a frame project, if videos should be split in frames
        fps: for a frame project, import videos with this frame rate
    """

    return {
        "processingParameters": {
            "shouldKeepNativeFrameRate": fps is None,
            "framesPlayedPerSecond": fps,
            "shouldUseNativeVideo": not as_frames,
        }
    }


@click.command(name="import")
@Arguments.files
@Options.api_key
@Options.endpoint
@Options.project_id
@from_csv(["external_id", "content"], [])
@click.option(
    "--frames",
    "as_frames",
    type=bool,
    default=False,
    is_flag=True,
    help=(
        "Only for a frame project, import videos as frames. "
        "The import time is longer with this option."
    ),
)
@click.option(
    "--fps", type=int, help="Only for a frame project, import videos with a specific frame rate"
)
@Options.verbose
@typechecked
# pylint: disable=too-many-arguments,too-many-locals
def import_assets(
    api_key: Optional[str],
    endpoint: Optional[str],
    project_id: str,
    files: Optional[Iterable[str]],
    csv_path: Optional[str],
    fps: Optional[int],
    as_frames: bool,
    verbose: bool,  # pylint: disable=unused-argument
):
    """
    Add assets into a project

    Files can be paths to files or to folders. You can provide several paths separated by spaces.

    If no Files are provided, --from-csv can be used to import
    assets from a CSV file with two columns:

    - `external_id`: external id of the asset.
    - `content`: paths to the asset file or a url hosting the asset.

    \b
    !!! Examples
        ```
        kili project import \\
            dir1/dir2/ dir1/dir3/test1.png \\
            --project-id <project_id>
        ```
        ```
        kili project import \\
            dir1/dir3/video.mp4 \\
            --project-id <project_id> \\
            --frames \\
            --fps 24
        ```
        ```
        kili project import \\
            --from-csv assets_list.csv \\
            --project-id <project_id> \\
            --frames \\
            --fps 24
        ```
    \b
    !!! warning "Unsupported imports"
        Currently, this command does not support:

        - the import of videos from local frames, rich text and time series assets
        - the import of assets with metadata or with a custom external_id

        For such imports, please use the `append_many_to_dataset` method in the Kili SDK.
    """
    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint)
    input_type = services.get_project_field(kili, project_id, "inputType")
    if input_type not in ("VIDEO_LEGACY", "VIDEO") and (fps is not None or as_frames is True):
        illegal_option = "fps and frames are"
        if not as_frames:
            illegal_option = "fps is"
        if fps is None:
            illegal_option = "frames is"
        raise ValueError(f"{illegal_option} only valid for a VIDEO project")

    check_exclusive_options(csv_path, files)

    if files:
        files_to_upload = get_file_paths_to_upload(
            list(files), file_check_function=None, verbose=False
        )
        if len(files_to_upload) == 0:
            raise ValueError(
                "No files to upload. "
                "Check that the paths exist and file types are compatible with the project"
            )
        external_ids = [get_external_id_from_file_path(path) for path in files_to_upload]

    elif csv_path is not None:
        row_dict = collect_from_csv(
            csv_path=csv_path,
            required_columns=["external_id", "content"],
            optional_columns=[],
            type_check_function=check_asset_type,
        )

        files_to_upload = [row["content"] for row in row_dict]
        external_ids = [row["external_id"] for row in row_dict]

        if len(files_to_upload) == 0:
            raise ValueError(f"No valid asset files or url were found in csv: {csv_path}")
    else:
        raise ValueError("You must provide either the file argument of the csv_path.")

    assets_to_import = [
        {
            "content": files_to_upload[i],
            "external_id": external_ids[i],
        }
        for i in range(len(files_to_upload))
    ]
    if input_type in ("VIDEO_LEGACY", "VIDEO"):
        json_metadata = generate_json_metadata(as_frames, fps)
        assets_to_import = [
            {**assets_to_import[i], "json_metadata": json_metadata}
            for i in range(len(assets_to_import))
        ]

    asset_import.import_assets(kili.auth, project_id, assets_to_import, False)



"""Plugins queries."""

import json
from datetime import datetime
from typing import List, Optional

from typeguard import typechecked

from kili.services.plugins import PluginUploader, get_logs, list_plugins

from ...authentication import KiliAuth


class QueriesPlugins:
    """Set of Plugins queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth: KiliAuth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def get_plugin_logs(
        self,
        project_id: str,
        plugin_name: str,
        start_date: Optional[datetime] = None,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
    ):
        # pylint: disable=line-too-long
        """Get paginated logs of a plugin on a project.
        Args:
            project_id: Identifier of the project
            plugin_name: Name of the plugin
            start_date: Datetime used to get the logs from, if not provided, it will be the plugin's creation date
            limit: Limit for pagination, if not provided, it will be 100
            skip: Skip for pagination, if not provided, it will be 0
        Returns:
            A result array which contains the logs of the plugin,
                or an error message.
        Examples:
            >>> kili.get_plugin_logs(project_id="my_project_id", plugin_name="my_plugin_name", start_date="1970/01/01")
        """

        plugin = {"project_id": project_id, "plugin_name": plugin_name}

        pretty_result = get_logs(self.auth, plugin, start_date, limit, skip)
        return json.dumps(pretty_result, sort_keys=True, indent=4)

    @typechecked
    def get_plugin_status(
        self,
        plugin_name: str,
        verbose: bool = True,
    ):
        """Update a plugin with new code.

        Args:
            plugin_name: Name of the plugin
            verbose: If false, minimal logs are displayed

        Returns:
            The status of the plugin if query was successful or an error message otherwise.

        Examples:
            >>> kili.get_plugin_status(plugin_name="my_plugin_name")
        """

        result = PluginUploader(self.auth, "", plugin_name, verbose).get_plugin_runner_status()
        return result

    # pylint: disable=dangerous-default-value
    @typechecked
    def list_plugins(
        self,
        fields: List[str] = [
            "name",
            "projectIds",
            "id",
            "createdAt",
            "updatedAt",
        ],
    ):
        # pylint: disable=line-too-long
        """List all plugins from your organization
        Args:
            fields: All the fields to request among the possible fields for the plugins
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#plugins) for all possible fields.

        Returns:
            A result array which contains all the plugins from your organization,
                or an error message.
        Examples:
            >>> kili.list_plugins()
            >>> kili.list_plugins(fields=['name'])
        """
        return list_plugins(self.auth, fields)



"""
Copy project implementation.
"""
import itertools
import logging
from typing import Dict, Optional

from kili import services
from kili.utils.tempfile import TemporaryDirectory
from kili.utils.tqdm import tqdm


class ProjectCopier:  # pylint: disable=too-few-public-methods
    """
    Class for copying an existing project.
    """

    FIELDS_PROJECT = ["title", "inputType", "description", "id"]
    FIELDS_JSON_INTERFACE = ["jsonInterface"]
    FIELDS_QUALITY_SETTINGS = [
        "consensusTotCoverage",
        "minConsensusSize",
        "useHoneyPot",
        "reviewCoverage",
    ]

    def __init__(self, kili) -> None:
        self.disable_tqdm = False
        self.kili = kili

    def copy_project(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        from_project_id: str,
        title: Optional[str],
        description: Optional[str],
        copy_json_interface: bool,
        copy_quality_settings: bool,
        copy_members: bool,
        copy_assets: bool,
        copy_labels: bool,
        disable_tqdm: bool,
    ) -> str:
        """
        Copy an existing project.
        """
        self.disable_tqdm = disable_tqdm

        logging.basicConfig()
        logger = logging.getLogger("kili.services.copy_project")
        logger.setLevel(logging.INFO)

        if not any(
            (copy_json_interface, copy_quality_settings, copy_members, copy_assets, copy_labels)
        ):
            raise ValueError("At least one element has to be copied.")

        if copy_labels:
            if not copy_json_interface:
                raise ValueError(
                    "`copy_json_interface` must be set to `True` for copying the source project"
                    " labels."
                )
            if not copy_assets:
                raise ValueError(
                    "`copy_assets` must be set to `True` for copying the source project labels."
                )
            if not copy_members:
                raise ValueError(
                    "`copy_members` must be set to `True` for copying the source project labels."
                )

        fields = self.FIELDS_PROJECT
        if copy_json_interface:
            fields = fields + self.FIELDS_JSON_INTERFACE
        if copy_quality_settings:
            fields = fields + self.FIELDS_QUALITY_SETTINGS

        src_project = services.get_project(self.kili, from_project_id, fields)

        new_project_title = title or self._generate_project_title(src_title=src_project["title"])

        new_project_description = description or ""

        json_interface = src_project["jsonInterface"] if copy_json_interface else {"jobs": {}}

        new_project_id = self.kili.create_project(
            input_type=src_project["inputType"],
            json_interface=json_interface,
            title=new_project_title,
            description=new_project_description,
        )["id"]
        logger.info(f"Creating new project with id: '{new_project_id}'")

        if copy_members:
            logger.info("Copying members...")
            self._copy_members(from_project_id, new_project_id)

        if copy_quality_settings:
            logger.info("Copying quality settings...")
            self._copy_quality_settings(new_project_id, src_project)

        if copy_assets:
            logger.info("Copying assets...")
            self._copy_assets(from_project_id, new_project_id)

        if copy_labels:
            logger.info("Copying labels...")
            self._copy_labels(from_project_id, new_project_id)

        return new_project_id

    def _generate_project_title(self, src_title: str) -> str:
        projects = self.kili.projects(
            fields=["title"], search_query=f"{src_title}%", disable_tqdm=True
        )
        proj_titles = {proj["title"] for proj in projects}
        new_title = f"{src_title} (copy)"
        i = 1
        while new_title in proj_titles:
            new_title = f"{src_title} (copy {i})"
            i += 1
        return new_title

    def _copy_members(self, from_project_id: str, new_project_id: str) -> None:
        members = self.kili.project_users(
            project_id=from_project_id,
            fields=["activated", "role", "user.email", "invitationStatus", "id"],
            disable_tqdm=True,
        )

        members = [
            memb
            for memb in members
            if memb["invitationStatus"] != "DEFAULT_ACCEPTED" and memb["activated"]
        ]

        for member in tqdm(members, disable=self.disable_tqdm):
            self.kili.append_to_roles(
                project_id=new_project_id,
                user_email=member["user"]["email"],
                role=member["role"],
            )

    def _copy_quality_settings(self, new_project_id: str, src_project: Dict) -> None:
        self.kili.update_properties_in_project(
            project_id=new_project_id,
            consensus_tot_coverage=src_project["consensusTotCoverage"],
            min_consensus_size=src_project["minConsensusSize"],
            use_honeypot=src_project["useHoneyPot"],
            review_coverage=src_project["reviewCoverage"],
        )

    # pylint: disable=too-many-locals
    def _copy_assets(self, from_project_id: str, new_project_id: str):
        """
        Copy assets from a project to another.

        Fetches assets by batch since `content` urls expire.
        """
        batch_size = 200

        nb_assets_to_copy = self.kili.count_assets(project_id=from_project_id)

        skip = 0
        with tqdm(total=nb_assets_to_copy, disable=self.disable_tqdm) as pbar:
            while skip < nb_assets_to_copy:
                with TemporaryDirectory() as tmp_dir:
                    assets = self.kili.assets(
                        project_id=from_project_id,
                        fields=[
                            "content",
                            "externalId",
                            "isHoneypot",
                            "jsonContent",
                            "jsonMetadata",
                        ],
                        skip=skip,
                        first=batch_size,
                        download_media=True,
                        local_media_dir=str(tmp_dir.resolve()),
                        disable_tqdm=True,
                    )

                    # cannot pass both content_array and json_content_array
                    # to append_many_to_dataset
                    # we sort and group by (asset["content"], asset["jsonContent"])
                    assets = sorted(
                        assets,
                        key=lambda asset: (asset["content"] != "", asset["jsonContent"] != ""),
                    )
                    assets_iterator = itertools.groupby(
                        assets,
                        key=lambda asset: (asset["content"] != "", asset["jsonContent"] != ""),
                    )

                    for key, group in assets_iterator:
                        has_content, has_jsoncontent = key
                        group = list(group)

                        content_array = (
                            [asset["content"] for asset in group] if has_content else None
                        )
                        external_id_array = [asset["externalId"] for asset in group]
                        is_honeypot_array = [asset["isHoneypot"] for asset in group]
                        json_content_array = (
                            [asset["jsonContent"] for asset in group] if has_jsoncontent else None
                        )
                        json_metadata_array = [asset["jsonMetadata"] for asset in group]

                        self.kili.append_many_to_dataset(
                            project_id=new_project_id,
                            content_array=content_array,
                            external_id_array=external_id_array,
                            is_honeypot_array=is_honeypot_array,
                            json_content_array=json_content_array,
                            json_metadata_array=json_metadata_array,
                            disable_tqdm=True,
                        )

                        skip += len(group)
                        pbar.update(skip)

    # pylint: disable=too-many-locals
    def _copy_labels(self, from_project_id: str, new_project_id: str):
        assets_new_project = self.kili.assets(
            new_project_id, fields=["externalId", "id"], disable_tqdm=True
        )
        assets_new_project_map = {asset["externalId"]: asset["id"] for asset in assets_new_project}

        members_new_project = self.kili.project_users(
            project_id=new_project_id, fields=["user.email", "user.id"], disable_tqdm=True
        )
        members_new_project_map = {
            member["user"]["email"]: member["user"]["id"] for member in members_new_project
        }

        nb_labels_to_copy = self.kili.count_labels(project_id=from_project_id)
        if nb_labels_to_copy == 0:
            return

        labels = self.kili.labels(
            project_id=from_project_id,
            fields=[
                "author.email",
                "jsonResponse",
                "secondsToLabel",
                "isLatestLabelForUser",
                "labelOf.externalId",
                "labelType",
                "modelName",
            ],
            disable_tqdm=True,
        )
        labels = [label for label in labels if label["isLatestLabelForUser"]]

        # `append_labels` does not take arrays for `model_name` and `label_type` arguments
        # we need to sort and group the labels by `model_name` and `label_type`
        # and upload the grouped labels by batch to `append_labels`
        labels = sorted(
            labels,
            key=lambda label: (label["labelType"], label["modelName"] is None, label["modelName"]),
        )
        labels_iterator = itertools.groupby(
            labels,
            key=lambda label: (label["labelType"], label["modelName"] is None, label["modelName"]),
        )

        for key, group in labels_iterator:
            label_type, _, model_name = key
            group = list(group)

            # map external id of source project asset to
            # internal id of new project corresponding asset
            # since both source_project and new_project have the same externalIds
            asset_id_array = [
                assets_new_project_map[label["labelOf"]["externalId"]] for label in group
            ]
            json_response_array = [label["jsonResponse"] for label in group]
            author_id_array = [members_new_project_map[label["author"]["email"]] for label in group]
            seconds_to_label_array = [label["secondsToLabel"] for label in group]

            self.kili.append_labels(
                asset_id_array=asset_id_array,
                json_response_array=json_response_array,
                author_id_array=author_id_array,
                seconds_to_label_array=seconds_to_label_array,
                model_name=model_name,
                label_type=label_type,
                disable_tqdm=False,
            )



"""
GraphQL Queries of Issues
"""

from typing import Optional

from kili.graphql import BaseQueryWhere, GraphQLQuery
from kili.types import Issue


class IssueWhere(BaseQueryWhere):
    """
    Tuple to be passed to the IssueQuery to restrict query
    """

    def __init__(
        self,
        project_id: Optional[str] = None,
    ):
        self.project_id = project_id
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK IssueWhere"""
        return {"project": {"id": self.project_id}}


class IssueQuery(GraphQLQuery):
    """Issue query."""

    TYPE = Issue

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL issues query
        """
        return f"""
    query issues($where: IssueWhere!, $first: PageSize!, $skip: Int!) {{
    data: issues(where: $where, first: $first, skip: $skip) {{
        {fragment}
    }}
    }}
    """

    COUNT_QUERY = """
    query countIssues($where: IssueWhere!) {
    data: countIssues(where: $where)
    }
    """



"""
Common code for the PASCAL VOC exporter.
"""

import warnings
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Dict, List, Tuple, Union
from xml.dom import minidom

from PIL import Image

from kili.orm import JobMLTask, JobTool
from kili.services.export.exceptions import (
    NoCompatibleJobError,
    NotCompatibleInputType,
    NotCompatibleOptions,
)
from kili.services.export.format.base import AbstractExporter
from kili.services.export.repository import AbstractContentRepository
from kili.utils.tqdm import tqdm


class VocExporter(AbstractExporter):
    """
    Common code for VOC exporter.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.with_assets:
            warnings.warn(
                "For an export to the Pascal VOC format, the download of assets cannot be disabled."
            )
        self.with_assets = True

    def _check_arguments_compatibility(self):
        """
        Checks if the export label format is compatible with the export options.
        """
        if self.single_file:
            raise NotCompatibleOptions(
                f"The label format {self.label_format} can not be exported into a single file."
            )
        if self.split_option != "merged":
            raise NotCompatibleOptions(
                "The current implementation only supports merged annotations."
            )

    def _check_project_compatibility(self) -> None:
        """
        Checks if the export label format is compatible with the project.
        """
        if self.project_input_type == "VIDEO":
            raise NotImplementedError("Export of annotations on videos is not supported yet.")
        if self.project_input_type != "IMAGE":
            raise NotCompatibleInputType(
                f"Project with input type '{self.project_input_type}' not compatible with Pascal"
                " VOC export format."
            )

        jobs = self.project_json_interface["jobs"]
        jobs = {
            job_name: job
            for job_name, job in jobs.items()
            if job["mlTask"] == JobMLTask.ObjectDetection
            and any(tool == JobTool.Rectangle for tool in job["tools"])
        }
        if not jobs:
            raise NoCompatibleJobError(
                f"Project needs at least one {JobMLTask.ObjectDetection} task with bounding boxes."
            )

    def process_and_save(self, assets: List[Dict], output_filename: Path) -> None:
        """
        Save the assets and annotations to a zip file in the Pascal VOC format.
        """
        # pylint: disable=too-many-locals, too-many-arguments
        self.logger.info("Exporting VOC format")

        labels_folder = self.base_folder / "labels"
        labels_folder.mkdir(parents=True)

        for asset in tqdm(assets, disable=self.disable_tqdm):
            _process_asset_for_job(self.content_repository, asset, labels_folder)

        self.create_readme_kili_file(self.export_root_folder)
        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)


def _process_asset_for_job(
    content_repository: AbstractContentRepository,
    asset: Dict,
    labels_folder: Path,
) -> None:
    # pylint: disable=too-many-locals, too-many-branches
    """
    Process an asset
    """
    frames = {}
    is_frame = False

    idx = 0
    if "jsonResponse" in asset["latestLabel"]:
        number_of_frames = len(asset["latestLabel"]["jsonResponse"])
        for idx in range(number_of_frames):
            if str(idx) in asset["latestLabel"]["jsonResponse"]:
                is_frame = True
                frame_asset = asset["latestLabel"]["jsonResponse"][str(idx)]
                frames[idx] = {"latestLabel": {"jsonResponse": frame_asset}}

    if is_frame:
        raise NotImplementedError("Export of annotations on videos is not supported yet.")

    if not frames:
        frames[-1] = asset

    content_frames = content_repository.get_content_frames_paths(asset)
    content_frame = content_frames[idx] if content_frames else asset["content"]

    filename = asset["externalId"]
    width, height = get_asset_dimensions(content_frame)
    for frame in frames.values():
        latest_label = frame["latestLabel"]
        json_response = latest_label["jsonResponse"]
        parameters = {"filename": f"{filename}.xml"}
        annotations = _convert_from_kili_to_voc_format(json_response, width, height, parameters)

        with open(labels_folder / f"{filename}.xml", "wb") as fout:
            fout.write(f"{annotations}\n".encode())


def get_asset_dimensions(file_path: Union[Path, str]) -> Tuple:
    """
    Get an asset width and height
    """
    image = Image.open(file_path)
    width, height = image.size
    return width, height


def _parse_annotations(
    response: Dict, xml_label: ET.Element, img_width: int, img_height: int
) -> None:
    # pylint: disable=too-many-locals
    for _, job_response in response.items():
        if "annotations" in job_response:
            annotations = job_response["annotations"]
            for annotation in annotations:
                vertices = annotation["boundingPoly"][0]["normalizedVertices"]
                categories = annotation["categories"]
                for category in categories:
                    annotation_category = ET.SubElement(xml_label, "object")
                    name = ET.SubElement(annotation_category, "name")
                    name.text = category["name"]
                    pose = ET.SubElement(annotation_category, "pose")
                    pose.text = "Unspecified"
                    truncated = ET.SubElement(annotation_category, "truncated")
                    truncated.text = "0"
                    difficult = ET.SubElement(annotation_category, "difficult")
                    difficult.text = "0"
                    occluded = ET.SubElement(annotation_category, "occluded")
                    occluded.text = "0"
                    bndbox = ET.SubElement(annotation_category, "bndbox")
                    x_vertices = [int(round(v["x"] * img_width)) for v in vertices]
                    y_vertices = [int(round(v["y"] * img_height)) for v in vertices]
                    xmin = ET.SubElement(bndbox, "xmin")
                    xmin.text = str(min(x_vertices))
                    xmax = ET.SubElement(bndbox, "xmax")
                    xmax.text = str(max(x_vertices))
                    ymin = ET.SubElement(bndbox, "ymin")
                    ymin.text = str(min(y_vertices))
                    ymax = ET.SubElement(bndbox, "ymax")
                    ymax.text = str(max(y_vertices))


def _provide_voc_headers(
    xml_label: ET.Element, img_width: int, img_height: int, parameters: Dict
) -> None:
    folder = ET.SubElement(xml_label, "folder")
    folder.text = parameters.get("folder", "")

    filename = ET.SubElement(xml_label, "filename")
    filename.text = parameters.get("filename", "")

    path = ET.SubElement(xml_label, "path")
    path.text = parameters.get("path", "")

    source = ET.SubElement(xml_label, "source")
    database = ET.SubElement(source, "database")
    database.text = "Kili Technology"

    size = ET.SubElement(xml_label, "size")
    width_xml = ET.SubElement(size, "width")
    width_xml.text = str(img_width)
    height_xml = ET.SubElement(size, "height")
    height_xml.text = str(img_height)
    depth = ET.SubElement(size, "depth")
    depth.text = parameters.get("depth", "3")

    segmented = ET.SubElement(xml_label, "segmented")
    segmented.text = None


def _convert_from_kili_to_voc_format(
    response: Dict, img_width: int, img_height: int, parameters: Dict
) -> str:
    xml_label = ET.Element("annotation")

    _provide_voc_headers(xml_label, img_width, img_height, parameters=parameters)

    _parse_annotations(response, xml_label, img_width, img_height)

    xmlstr = minidom.parseString(ET.tostring(xml_label)).toprettyxml(indent="   ")

    return xmlstr



"""Project user queries."""

from typing import Dict, Iterable, List, Optional

from typeguard import typechecked

from kili.graphql import QueryOptions
from kili.graphql.operations.project_user.queries import (
    ProjectUserQuery,
    ProjectUserWhere,
)


class QueriesProjectUser:
    """Set of ProjectUser queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value,invalid-name
    @typechecked
    def project_users(
        self,
        project_id: str,
        email: Optional[str] = None,
        id: Optional[str] = None,  # pylint: disable=redefined-builtin
        organization_id: Optional[str] = None,
        fields: List[str] = [
            "activated",
            "id",
            "role",
            "starred",
            "user.email",
            "user.id",
        ],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Return project users (possibly with their KPIs) that match a set of criteria


        Args:
            project_id: Identifier of the project
            email: Email of the user
            organization_id: Identifier of the user's organization
            fields: All the fields to request among the possible fields for the projectUsers
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#projectuser) for all possible fields.
            first: Maximum number of users to return
            skip: Number of project users to skip
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the project users is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            ```
            # Retrieve consensus marks of all users in project
            >>> kili.project_users(project_id=project_id, fields=['consensusMark', 'user.email'])
            ```
        """
        where = ProjectUserWhere(
            project_id=project_id, email=email, _id=id, organization_id=organization_id
        )
        options = QueryOptions(disable_tqdm, first, skip, as_generator)
        return ProjectUserQuery(self.auth.client)(where, fields, options)

    # pylint: disable=invalid-name
    @typechecked
    def count_project_users(
        self,
        project_id: str,
        email: Optional[str] = None,
        id: Optional[str] = None,  # pylint: disable=redefined-builtin
        organization_id: Optional[str] = None,
    ) -> int:
        """
        Counts the number of projects and their users that match a set of criteria

        Args:
            email: Email of the user
            organization_id: Identifier of the user's organization
            project_id: Identifier of the project

        Returns:
            The number of project users with the parameters provided
        """
        where = ProjectUserWhere(
            project_id=project_id, email=email, _id=id, organization_id=organization_id
        )
        return ProjectUserQuery(self.auth.client).count(where)



"""
GraphQL Queries of Users
"""


from typing import Optional

from kili.graphql import BaseQueryWhere, GraphQLQuery
from kili.types import User


class UserWhere(BaseQueryWhere):
    """
    Tuple to be passed to the UserQuery to restrict query
    """

    # pylint: disable=too-many-arguments,too-many-locals,too-many-instance-attributes

    def __init__(
        self,
        api_key: Optional[str] = None,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
    ):
        self.api_key = api_key
        self.email = email
        self.organization_id = organization_id
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK UserWhere"""
        return {
            "apiKey": self.api_key,
            "email": self.email,
            "organization": {"id": self.organization_id},
        }


class UserQuery(GraphQLQuery):
    """User query."""

    TYPE = User

    @staticmethod
    def query(fragment):
        """
        Return the GraphQL users query
        """
        return f"""
    query users($where: UserWhere!, $first: PageSize!, $skip: Int!) {{
    data: users(where: $where, first: $first, skip: $skip) {{
        {fragment}
    }}
    }}
    """

    COUNT_QUERY = """
    query countUsers($where: UserWhere!) {
    data: countUsers(where: $where)
    }
    """


GQL_ME = """
query Me {
  data: me {
    id
    email
  }
}
"""



"""Project mutations."""
from typing import Optional

from typeguard import typechecked

from kili.authentication import KiliAuth
from kili.services.plugins import (
    PluginUploader,
    WebhookUploader,
    activate_plugin,
    deactivate_plugin,
    delete_plugin,
)


class MutationsPlugins:
    """Set of Plugins mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth: KiliAuth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @typechecked
    def upload_plugin(
        self,
        plugin_path: Optional[str] = None,
        plugin_name: Optional[str] = None,
        verbose: bool = True,
        **kwargs
    ):
        # pylint: disable=line-too-long
        """Uploads a plugin.

        Args:
            plugin_path : Path to your plugin. Either a folder containing a main.py (mandatory) and a requirements.txt (optional) or a .py file
            plugin_name: name of your plugin, if not provided, it will be the name from your file
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.upload_plugin(plugin_path="./path/to/my/file.py")
        """

        if kwargs.get("file_path"):
            raise TypeError(
                '"file_path" has been deprecated for "plugin_path", please use "plugin_path" instead'
            )

        if not plugin_path:
            raise TypeError('"plugin_path is nullish, please provide a value')

        return PluginUploader(self.auth, plugin_path, plugin_name, verbose).create_plugin()

    @typechecked
    def create_webhook(
        self,
        webhook_url: str,
        plugin_name: str,
        header: Optional[str] = None,
        verbose: bool = True,
    ):
        # pylint: disable=line-too-long
        """Create a webhook linked to Kili's events.

        Args:
            webhook_url: URL receiving post requests on events on Kili
            plugin_name: name of your plugin
            header: Authorization header to access the routes
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.create_webhook(webhook_url='https://my-custom-url-publicly-accessible/', plugin_name='my webhook', header='...')
        """

        return WebhookUploader(
            self.auth, webhook_url, plugin_name, header, verbose
        ).create_webhook()

    @typechecked
    def update_webhook(
        self,
        new_webhook_url: str,
        plugin_name: str,
        new_header: Optional[str] = None,
        verbose: bool = True,
    ):
        # pylint: disable=line-too-long
        """Update a webhook linked to Kili's events.

        Args:
            new_webhook_url: New URL receiving post requests on events on Kili
            plugin_name: name of your plugin
            new_header: Authorization header to access the routes
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_webhook(webhook_url='https://my-custom-url-publicly-accessible/', plugin_name='my webhook', header='...')
        """

        return WebhookUploader(
            self.auth, new_webhook_url, plugin_name, new_header, verbose
        ).update_webhook()

    @typechecked
    def activate_plugin_on_project(
        self,
        plugin_name: str,
        project_id: str,
    ):
        # pylint: disable=line-too-long
        """Activates a plugin on a project.

        Args:
            plugin_name: Name of the plugin
            project_id: Identifier of the project
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.activate_plugin_on_project(plugin_name="my_plugin_name", project_id="my_project_id")
        """

        return activate_plugin(self.auth, plugin_name, project_id)

    @typechecked
    def deactivate_plugin_on_project(
        self,
        plugin_name: str,
        project_id: str,
    ):
        # pylint: disable=line-too-long
        """Activates a plugin on a project.

        Args:
            plugin_name: Name of the plugin
            project_id: Identifier of the project
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.deactivate_plugin_on_project(plugin_name="my_plugin_name", project_id="my_project_id")
        """

        return deactivate_plugin(self.auth, plugin_name, project_id)

    @typechecked
    def delete_plugin(
        self,
        plugin_name: str,
    ):
        # pylint: disable=line-too-long
        """Deletes a plugin.

        Args:
            plugin_name: Name of the plugin
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.delete_plugin(plugin_name="my_plugin_name")
        """

        return delete_plugin(self.auth, plugin_name)

    @typechecked
    def update_plugin(
        self,
        plugin_path: Optional[str] = None,
        plugin_name: Optional[str] = None,
        verbose: bool = True,
        **kwargs
    ):
        """Update a plugin with new code.

        Args:
            plugin_path : Path to your plugin. Either:
             - a folder containing a main.py (mandatory) and a requirements.txt (optional)
             - a .py file
            plugin_name: Name of the plugin
            verbose: If false, minimal logs are displayed

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.

        Examples:
            >>> kili.update_plugin(plugin_name="my_plugin_name")
        """

        if kwargs.get("file_path"):
            raise TypeError(
                """ "file_path" has been deprecated for "plugin_path",
                please use "plugin_path" instead"""
            )

        if not plugin_path:
            raise TypeError('"plugin_path is nullish, please provide a value')

        if not plugin_name:
            raise TypeError('"plugin_name is nullish, please provide a value')

        return PluginUploader(self.auth, plugin_path, plugin_name, verbose).update_plugin()



"""Issue queries."""

import warnings
from typing import Iterable, List, Optional

from typeguard import typechecked

from kili.graphql import QueryOptions
from kili.graphql.operations.issue.queries import IssueQuery, IssueWhere


class QueriesIssue:
    """Set of Issue queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @typechecked
    def issues(
        self,
        project_id: str,
        fields: List[str] = [
            "id",
            "createdAt",
            "hasBeenSeen",
            "issueNumber",
            "status",
            "type",
        ],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Iterable[dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of issues that match a set of criteria.

        Args:
            project_id: Project ID the issue belongs to.
            fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#issue) for all possible fields.
            first: Maximum number of issues to return.
            skip: Number of issues to skip (they are ordered by their date of creation, first to last).
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the issues is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            >>> kili.issues(project_id=project_id, fields=['author.email']) # List all issues of a project and their authors
        """
        where = IssueWhere(project_id=project_id)
        options = QueryOptions(disable_tqdm, first, skip, as_generator)
        return IssueQuery(self.auth.client)(where, fields, options)

    @typechecked
    def count_issues(self, project_id: Optional[str] = None) -> int:
        """Count and return the number of api keys with the given constraints.

        Args:
            project_id: Project ID the issue belongs to.

        Returns:
            The number of issues with the parameters provided

        """
        if not project_id:
            warnings.warn(
                "It is now required to provide a project_id when calling count_issues. This change"
                " will be enforced from 01/02/2023",
                DeprecationWarning,
            )
        where = IssueWhere(project_id=project_id)
        return IssueQuery(self.auth.client).count(where)



"""
GraphQL module
"""

from abc import ABC, abstractmethod
from typing import Dict, Iterable, List, NamedTuple, Optional, Type

from tqdm import tqdm

from kili.helpers import format_result, fragment_builder
from kili.utils.pagination import api_throttle

from .graphql_client import GraphQLClient


class QueryOptions(NamedTuple):
    """Options when calling GraphQLQuery from the SDK"""

    disable_tqdm: Optional[bool]
    first: Optional[int] = None
    skip: int = 0
    as_generator: bool = False


class BaseQueryWhere(ABC):
    """
    Abtsract class for defining the where payload to send in a graphQL query
    """

    def __init__(self):
        self._graphql_payload = self.graphql_where_builder()

    @abstractmethod
    def graphql_where_builder(self) -> Dict:
        """Build the GraphQL where payload sent in the resolver from the
        arguments given to the where class
        """
        raise NotImplementedError

    @property
    def graphql_payload(self):
        """where payload to send in the graphQL query"""
        return self._graphql_payload


class GraphQLQuery(ABC):
    """
    Query class for querying Kili objects.
    It factorizes code for executing paginated queries
    """

    def __init__(
        self,
        client: GraphQLClient,
    ):
        self.client = client

    @staticmethod
    @abstractmethod
    def query(fragment: str) -> str:
        """Return the GraphQL object query to be sent, according to the fragment to return.

        Args:
            fragment: the fragment to return in the query

        Return:
            The query to be sent to graphQL
        """
        raise NotImplementedError

    COUNT_QUERY: str = NotImplemented

    TYPE: Type = NotImplemented

    # to be implemented when adding the asset query with the new architecture
    # @staticmethod
    # @abstractmethod
    # def post_call_process():

    def __call__(
        self,
        where: BaseQueryWhere,
        fields: List[str],
        options: QueryOptions,
    ) -> Iterable[Dict]:
        """Query objects of the specified type"""
        fragment = fragment_builder(fields, self.TYPE)
        query = self.query(fragment)

        result_gen = self.execute_query_from_paginated_call(query, where, options)
        if options.as_generator:
            return result_gen
        return list(result_gen)

    def count(self, where: BaseQueryWhere):
        """Count the number of objects matching the given where payload"""
        payload = {"where": where.graphql_payload}
        count_result = self.client.execute(self.COUNT_QUERY, payload)
        return format_result("data", count_result, int)

    def get_number_of_elements_to_query(self, where: BaseQueryWhere, options: QueryOptions):
        """Return the total number of element to query for one query.
        It uses both the argument first given by the user
        and the total number of available objects obtained with a graphQL count query
        """
        first = options.first
        skip = options.skip
        count_rows_available = self.count(where)
        count_objects_queried = max(count_rows_available - skip, 0)
        if first is None:
            return count_objects_queried
        return min(count_objects_queried, first)

    def execute_query_from_paginated_call(
        self, query: str, where: BaseQueryWhere, options: QueryOptions
    ):
        """
        Builds a row generator from paginated calls.

        Args:
            query: The object query to execute and to send to graphQL, in string format
            where_payload: The where payload to be sent to graphQL,
                as a value of the 'where' key in the global payload
            options: The query options
        """
        if options.as_generator and not options.disable_tqdm:
            options._replace(disable_tqdm=True)
        total_rows_queried = self.get_number_of_elements_to_query(where, options)
        batch_size = min(100, options.first or 100)

        if total_rows_queried == 0:
            yield from ()
        else:
            with tqdm(total=total_rows_queried, disable=options.disable_tqdm) as pbar:
                count_rows_retrieved = 0
                while count_rows_retrieved < total_rows_queried:
                    skip = count_rows_retrieved + options.skip
                    payload = {"where": where.graphql_payload, "skip": skip, "first": batch_size}
                    rows = api_throttle(self.client.execute)(query, payload)
                    rows = format_result("data", rows)

                    if rows is None or len(rows) == 0:
                        break

                    # to be implemented when adding the asset query with the new architecture
                    # if self.post_call_process is not None:
                    #     rows = self.post_call_process(rows)

                    for row in rows:
                        yield row

                    count_rows_retrieved += len(rows)
                    pbar.update(len(rows))



"""
Api key queries
"""

from typing import Dict, Iterable, List, Optional

from typeguard import typechecked

from kili.graphql import QueryOptions
from kili.graphql.operations.api_key.queries import APIKeyQuery, APIKeyWhere


class QueriesApiKey:
    """
    Set of ApiKey queries
    """

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @typechecked
    def api_keys(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
        skip: int = 0,
        fields: List[str] = ["id", "name", "createdAt", "revoked"],
        first: Optional[int] = None,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of API keys that match a set of constraints.

        !!! info
            You can only query your own API keys

        Args:
            api_key_id: Identifier of the API key to retrieve.
            user_id: Identifier of the user.
            api_key: Value of the API key.
            skip: Number of assets to skip (they are ordered by their date of creation, first to last).
            fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#apikey) for all possible fields.
            first: Maximum number of API keys to return.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the API key is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.


        Examples:
            >>> kili.api_keys(user_id=user_id)
            >>> kili.api_keys(api_key=api_key)
            >>> kili.api_keys(api_key=api_key, as_generator=False)
        """
        where = APIKeyWhere(api_key_id=api_key_id, user_id=user_id, api_key=api_key)
        options = QueryOptions(disable_tqdm, first, skip, as_generator)
        return APIKeyQuery(self.auth.client)(where, fields, options)

    @typechecked
    def count_api_keys(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
    ) -> int:
        """Count and return the number of api keys with the given constraints.

        Args:
            api_key_id: Identifier of the API key to retrieve.
            user_id: Identifier of the user.
            api_key: Value of the api key.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            >>> kili.count_api_keys(user_id=user_id)
            3
            >>> kili.count_api_keys(api_key=api_key)
            1
        """
        where = APIKeyWhere(api_key_id=api_key_id, user_id=user_id, api_key=api_key)
        return APIKeyQuery(self.auth.client).count(where)



"""
Common code for the Kili exporter.
"""

import json
from pathlib import Path
from typing import List

from kili.orm import Asset
from kili.services.export.format.base import AbstractExporter


class KiliExporter(AbstractExporter):
    """
    Common code for Kili exporters.
    """

    def _check_arguments_compatibility(self):
        """
        Checks if the export label format is compatible with the export options.
        """

    def _check_project_compatibility(self) -> None:
        """
        Checks if the export label format is compatible with the project.
        """

    def _save_assets_export(self, assets: List[Asset], output_filename: Path):
        """
        Save the assets to a file and return the link to that file
        """
        self.logger.info("Exporting to kili format...")

        if self.single_file:
            project_json = json.dumps(assets, sort_keys=True, indent=4)
            with (self.base_folder / "data.json").open("wb") as output_file:
                output_file.write(project_json.encode("utf-8"))
        else:
            labels_folder = self.base_folder / "labels"
            labels_folder.mkdir(parents=True)
            for asset in assets:
                external_id = asset["externalId"].replace(" ", "_")
                asset_json = json.dumps(asset, sort_keys=True, indent=4)
                with (labels_folder / f"{external_id}.json").open("wb") as output_file:
                    output_file.write(asset_json.encode("utf-8"))
        self.create_readme_kili_file(self.export_root_folder)

        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)

    def process_and_save(self, assets: List[Asset], output_filename: Path):
        """
        Extract formatted annotations from labels and save the json in the buckets.
        """
        clean_assets = self.process_assets(assets, self.label_format)
        return self._save_assets_export(
            clean_assets,
            output_filename,
        )

    @property
    def images_folder(self) -> Path:
        """
        Export images folder
        """
        return self.base_folder / "assets"



"""
Helpers for the services
"""
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, TypeVar

from typeguard import typechecked

from kili.exceptions import NotFound
from kili.services.exceptions import (
    NotEnoughArgumentsSpecifiedError,
    TooManyArgumentsSpecifiedError,
)
from kili.services.project import get_project_field

PathLike = TypeVar("PathLike", Path, str)


def check_exclusive_options(
    csv_path: Optional[PathLike], files: Optional[Iterable[PathLike]]
) -> None:
    """Forbid mutual use of options and argument(s)"""
    if files is not None:
        files = list(files)

    if (csv_path is not None) + (files is not None and len(files) > 0) > 1:
        raise TooManyArgumentsSpecifiedError(
            "An explicit list of files and a CSV file containing label files"
            " can't be specified at the same time"
        )

    if (csv_path is not None) + (files is not None and len(files) > 0) == 0:
        raise NotEnoughArgumentsSpecifiedError(
            "You must either provide an explicit list of files or a CSV file containing a file list"
        )


def get_external_id_from_file_path(path: PathLike) -> str:
    """Return external_id from file's path
    ex: 'tree/leaf/file_name.txt- -> file_name
    """
    file_path = Path(path).parts[-1]
    return ".".join(file_path.split(".")[:-1])


def is_target_job_in_json_interface(kili, project_id: str, target_job_name: str):
    """Tell if the target job id is defined in the project's JSON interface"""
    json_interface = get_project_field(kili, project_id, "jsonInterface")
    return target_job_name in json_interface["jobs"]


@typechecked
def infer_ids_from_external_ids(
    kili, asset_external_ids: List[str], project_id: str
) -> Dict[str, str]:
    """
    Infer asset ids from their external ids and project Id.

    Returns a dict that maps external ids to internal ids.

    Args:
        asset_id: asset id
        external_id: external id
        project_id: project id
    """
    assets = kili.assets(
        external_id_contains=asset_external_ids,
        project_id=project_id,
        fields=["id", "externalId"],
        disable_tqdm=True,
    )
    id_map: Dict[str, str] = {}
    for asset in assets:
        id_map[asset["externalId"]] = asset["id"]

    if len(id_map) < len(set(asset_external_ids)):
        assets_not_found = [
            external_id for external_id in asset_external_ids if external_id not in id_map
        ]
        raise NotFound(
            f"The assets whose external_id are: {assets_not_found} have not been found in the"
            f" project of Id {project_id}"
        )
    if len(id_map) > len(set(asset_external_ids)):
        raise NotFound(
            "Several assets have been found for the same external_id. Please consider using asset"
            " ids instead."
        )
    return id_map


def assert_all_arrays_have_same_size(arrays: List[Optional[List[Any]]], raise_error=True):
    """Assert that all given arrays have the same size if they are not None"""
    sizes_arrays = {len(array) for array in arrays if array is not None}
    if len(sizes_arrays) > 1:
        if raise_error:
            raise ValueError("All arrays should have the same length")
        return False
    return True
